<!-- "Head.html" -->
<meta http-equiv="Content-type" content="text/html; charset=utf-8">
<meta name="viewport" content="width=device-width">
{% if csettings.Mezzanine == True %}
  {% load pages_tags mezzanine_tags i18n future %}
{% endif %}
